"""
__version__.py
~~~~~~~~~~~~~~

Information about the current version of the package.
"""

__title__ = "emergefs"
__description__ = "Emerge Computational Filesystem."
__version__ = "0.0.5"
__author__ = "Darren Govoni"
__author_email__ = "darren@elasticcode.ai"
__license__ = "MIT License"
__url__ = "https://github.com/radiantone/emerge"
